"""ファイルストレージモジュール."""
import json
import os

from .base import TemplateStorage
from ..core.defines import PromptTemplate


class FileStorage(TemplateStorage):
    """ファイルにテンプレートを保存するストレージ."""

    def __init__(self, folder: str) -> None:
        """コンストラクタ."""
        self.folder = folder

    def _make_path(self, name: str, version: str) -> str:
        """ファイルパスを生成する."""
        return os.path.join(self.folder, name, f'{name}_{version}.json')

    def save_template(self, template: PromptTemplate) -> None:
        """テンプレートを保存する."""
        filename = self._make_path(template.name, template.version)
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, 'w') as f:
            json.dump(template.model_dump(), f)

    def load_template(self, name: str, version: str) -> PromptTemplate:
        """指定された名前とバージョンのテンプレートをロードする."""
        filename = self._make_path(name, version)
        with open(filename, 'r') as f:
            template = PromptTemplate.model_validate(json.load(f))
        return template

    def list_templates(self) -> list[str]:
        """全テンプレートをリストアップする."""
        return os.listdir(self.folder)

    def list_versions(self, name) -> list[str]:
        """指定された名前のテンプレートの全バージョンをリストアップする."""
        def split_name(filename):
            return filename[:-5].split('_')[1]

        folder = os.path.join(self.folder, name)
        return [split_name(f) for f in os.listdir(folder)]



"""LLMAdapterモジュール."""
from abc import abstractmethod

from ..core.defines import BuiltPrompt


class LLMAdapter:
    """LLMとの通信を行うための抽象クラス."""

    @abstractmethod
    def call_llm(self, prompt: BuiltPrompt) -> tuple[str, int, int]:
        """
        LLMを呼び出す.

        Function Callを使って、返答をパラメータに当てはめる際に使用する。
        """
        raise NotImplementedError()

    @abstractmethod
    def invoke_llm(self, prompt: BuiltPrompt, stream=False) -> str:
        """
        LLMを呼び出す.

        LLMの返答をそのまま返す。
        """
        raise NotImplementedError()



"""メモリストレージ定義モジュール."""
from .base import TemplateStorage
from ..core.defines import PromptTemplate


class MemoryStorage(TemplateStorage):
    """メモリ上にテンプレートを保存するストレージ."""

    def __init__(self):
        """コンストラクタ."""
        self._templates = {}

    def save_template(self, template: PromptTemplate) -> None:
        """テンプレートを保存する."""
        if template.name not in self._templates:
            self._templates[template.name] = {}
        if template.version in self._templates[template.name]:
            raise ValueError(f'{template.name} {template.version} is already exists.')
        self._templates[template.name][template.version] = template

    def load_template(self, name: str, version: str) -> PromptTemplate:
        """指定された名前とバージョンのテンプレートをロードする."""
        if name not in self._templates:
            raise ValueError(f'{name} is not found.')
        if version not in self._templates[name]:
            raise ValueError(f'{name} {version} is not found.')
        return self._templates[name][version]

    def list_templates(self) -> list[str]:
        """全テンプレートをリストアップする."""
        return list(self._templates.keys())

    def list_versions(self, name) -> list[str]:
        """指定された名前のテンプレートの全バージョンをリストアップする."""
        if name not in self._templates:
            raise ValueError(f'{name} is not found')
        return list(self._templates[name].keys())



"""RAG処理を行うモジュール."""
import json
import os

from .base import Retriever


class LocalVectorProcessor:
    """ベクトルデータから対象データを抽出処理を行うクラス."""

    def __init__(self, model, distance=60) -> None:
        """コンストラクタ."""
        self.model = model
        self.distance = distance
        import faiss
        import numpy
        self.faiss = faiss
        self.np = numpy

    def query(self, query: str, max_result: int) -> list[dict[str, str]]:
        """RAGクエリを実行する."""
        embed = self.np.array([self.model.encode(query)])
        D, idx = self.index.search(embed, max_result)
        results = [{
            'Q': self.map[i]['Q'],
            'A': self.map[i]['A'],
            'distance': str(d)
        } for i, d in zip(idx[0], D[0]) if d < self.distance]
        return results

    def load_index(self, home: str, name: str) -> bool:
        """インデックスをロードする."""
        index_file = f'{home}/{name}.db'
        if not os.path.exists(index_file):
            return False
        self.index = self.faiss.read_index(index_file)
        with open(f'{home}/{name}.map', 'r') as f:
            self.map = json.load(f)
        return True

    def make_index(self, home: str, dataset: str) -> None:
        """インデックスを作成する."""
        inp = f'{home}/{dataset}'
        with open(inp, 'r') as f:
            self.map = json.load(f)

        embeds = self.np.array([self.model.encode(d['Q']) for d in self.map])
        self.index = self.faiss.IndexFlatL2(embeds.shape[1], )
        self.index.add(embeds)

    def save_index(self, home: str, name: str) -> None:
        """インデックスを保存する."""
        index_file = f'{home}/{name}.db'
        self.faiss.write_index(self.index, index_file)
        with open(f'{home}/{name}.map', 'w') as f:
            json.dump(self.map, f)


class VectorQuery(Retriever):
    """RAGクエリークラス."""

    def __init__(self, processor: LocalVectorProcessor, query: str, max_results: int = 5) -> None:
        """コンストラクタ."""
        self.processor = processor
        self.query_str = query
        self.max_results = max_results

    def query(self) -> list[str]:
        """クエリを実行する."""
        results = self.processor.query(self.query_str, self.max_results)
        return [r['A'] for r in results]



"""OpenAI APIモジュール."""
from openai import OpenAI

from .adapter import LLMAdapter
from ..core.defines import BuiltPrompt
from ..core.defines import Function
from ..core.defines import ObjectProperty
from ..core.defines import ArrayProperty
from ..core.defines import StringProperty

client = OpenAI()


class OpenAIGPT4oMini(LLMAdapter):

    def __init__(self) -> None:
        """コンストラクタ."""
        # FIXME その他パラメータをどのように設定するか？
        pass

    def _build_message(self, message) -> dict:
        """メッセージを構築する."""
        if message.content_type == 'text':
            return message.content

        return {
            'type': message.content_type,
            message.content_type: message.content
        }

    def _build_properties(self, properties) -> dict:
        """プロパティを構築する."""
        props = {}
        for prop in properties:
            if isinstance(prop, ObjectProperty):
                props[prop.name] = {
                    'type': 'object',
                    'properties': self._build_properties(prop.properties)
                }
            elif isinstance(prop, ArrayProperty):
                props[prop.name] = {
                    'type': 'array',
                    'items': self._build_properties([prop.items])
                }
            elif isinstance(prop, StringProperty):
                props[prop.name] = {
                    'type': 'string',
                    'description': prop.description
                }
        return props

    def _build_function(self, func: Function) -> dict:
        """ツール設定を構築する."""
        return {
            'name': func.name,
            'description': func.description,
            'parameters': {
                'type': 'object',
                'properties': self._build_properties(func.properties),
                'additionalProperties': False
            }
        }

    def build_prompt(self, prompt: BuiltPrompt) -> dict:
        data = {
            'model': 'gpt-4o-mini',
        }
        if prompt.system:
            data['messages'] = [{
                'role': 'system',
                'content': prompt.system
            }]
        else:
            data['messages'] = []

        data['messages'].extend([{
                'role': message.role,
                'content': self._build_message(message)
            } for message in prompt.messages])

        if prompt.function_calling_config:
            data['tools'] = [{
                'type': 'function',
                'function': self._build_function(func)
            } for func in prompt.function_calling_config.functions]
            if prompt.function_calling_config.function_call:
                data['tool_choice'] = 'required'
        return data

    def call_llm(self, prompt: BuiltPrompt) -> tuple[str, int, int]:
        """LLMを呼び出す."""
        data = self.build_prompt(prompt)
        result = client.chat.completions.create(**data)

        if prompt.function_calling_config:
            output = result.choices[0].message.tool_calls[0].function.arguments
        else:
            output = result.choices[0].message.content
        input_token = result.usage.prompt_tokens
        output_tokens = result.usage.completion_tokens

        return output, input_token, output_tokens

    def invoke_llm(self, prompt: BuiltPrompt, stream=False) -> str:
        data = self.build_prompt(prompt)
        if stream:
            data['stream'] = True
        return client.chat.completions.create(**data)



"""Promptia Core Modules."""
from .core.manager import Promptia
from .core.defines import PromptTemplate
from .core.defines import Message
from .core.defines import BuiltPrompt
from .core.defines import FunctionCallingConfig
from .core.defines import Function
from .core.defines import Property
from .core.defines import ObjectProperty
from .core.defines import ArrayProperty
from .core.defines import StringProperty
from .retrievers.base import Retriever
from .retrievers.vector_retriever import VectorQuery
from .retrievers.vector_retriever import LocalVectorProcessor
from .storages.base import TemplateStorage
from .storages.file_storage import FileStorage
from .storages.memory_storage import MemoryStorage
from .storages.s3_storage import S3Storage

__version__ = '0.1.0'
__all__ = [
    'Promptia',
    'PromptTemplate',
    'Message',
    'BuiltPrompt',
    'FunctionCallingConfig',
    'Function',
    'Property',
    'ObjectProperty',
    'ArrayProperty',
    'StringProperty',
    'Retriever',
    'VectorQuery',
    'LocalVectorProcessor',
    'TemplateStorage',
    'MemoryStorage',
    'FileStorage',
    'S3Storage'
]



"""Azure経由でOpenAI APIを利用するモジュール."""
import os

from openai import AzureOpenAI

from .adapter import LLMAdapter
from ..core.defines import BuiltPrompt
from ..core.defines import Function
from ..core.defines import ObjectProperty
from ..core.defines import ArrayProperty
from ..core.defines import StringProperty

client = AzureOpenAI(
    api_version=os.getenv('AZURE_API_VERSION'),
    azure_endpoint=os.getenv('AZURE_ENDPOINT'),
)


class AzureGPT4oMini(LLMAdapter):

    def __init__(self) -> None:
        """コンストラクタ."""
        # FIXME その他パラメータをどのように設定するか？
        pass

    def _build_message(self, message) -> dict:
        """メッセージを構築する."""
        if message.content_type == 'text':
            return message.content

        return {
            'type': message.content_type,
            message.content_type: message.content
        }

    def _build_properties(self, properties) -> dict:
        """プロパティを構築する."""
        props = {}
        for prop in properties:
            if isinstance(prop, ObjectProperty):
                props[prop.name] = {
                    'type': 'object',
                    'properties': self._build_properties(prop.properties)
                }
            elif isinstance(prop, ArrayProperty):
                props[prop.name] = {
                    'type': 'array',
                    'items': self._build_properties([prop.items])
                }
            elif isinstance(prop, StringProperty):
                props[prop.name] = {
                    'type': 'string',
                    'description': prop.description
                }
        return props

    def _build_function(self, func: Function) -> dict:
        """ツール設定を構築する."""
        return {
            'name': func.name,
            'description': func.description,
            'parameters': {
                'type': 'object',
                'properties': self._build_properties(func.properties),
                'additionalProperties': False
            }
        }

    def call_llm(self, prompt: BuiltPrompt) -> tuple[str, int, int]:
        """LLMを呼び出す."""

        data = {
            'model': 'gpt-4o-mini',
        }
        if prompt.system:
            data['messages'] = [{
                'role': 'system',
                'content': prompt.system
            }]
        else:
            data['messages'] = []

        data['messages'].extend([{
                'role': message.role,
                'content': self._build_message(message)
            } for message in prompt.messages])

        if prompt.function_calling_config:
            data['tools'] = [{
                'type': 'function',
                'function': self._build_function(func)
            } for func in prompt.function_calling_config.functions]
            if prompt.function_calling_config.function_call:
                data['tool_choice'] = 'required'

        result = client.chat.completions.create(**data)

        if prompt.function_calling_config:
            output = result.choices[0].message.tool_calls[0].function.arguments
        else:
            output = result.choices[0].message.content
        input_token = result.usage.prompt_tokens
        output_tokens = result.usage.completion_tokens

        return output, input_token, output_tokens



"""取得処理系モジュール."""
from abc import abstractmethod


class Retriever:
    """リトリーバクラス."""

    @abstractmethod
    def query(self) -> list[str]:
        """クエリを実行する."""
        raise NotImplementedError()



"""Promptiaのメインロジックを実装するモジュール."""
from jinja2 import Template

from .defines import BuiltPrompt
from .defines import PromptTemplate
from ..retrievers.base import Retriever


class Promptia:
    """Promptiaのメインロジックを実装するクラス."""

    def __init__(self) -> None:
        """コンストラクタ."""
        pass

    def build(self, template: PromptTemplate, values: dict[str, str | Retriever]) -> BuiltPrompt:
        """プロンプトを構築する."""
        processed_values = {}

        for key, value in values.items():
            if isinstance(value, Retriever):
                processed_values[key] = value.query()
            else:
                processed_values[key] = value

        tmpl = Template(source=template.system)
        rendered_content = tmpl.render(processed_values)
        return BuiltPrompt(
            system=rendered_content,
            messages=template.messages,
            function_calling_config=template.function_calling_config
        )



"""テンプレートの永続化を行うためのインターフェースを提供するモジュール."""
from abc import abstractmethod

from ..core.defines import PromptTemplate


class TemplateStorage:
    """テンプレートの永続化を行うためのインターフェース."""

    @abstractmethod
    def save_template(self, template: PromptTemplate) -> None:
        """テンプレートを保存する."""
        raise NotImplementedError()

    @abstractmethod
    def load_template(self, name: str, version: str) -> PromptTemplate:
        """指定された名前とバージョンのテンプレートをロードする."""
        raise NotImplementedError()

    @abstractmethod
    def list_templates(self) -> list[str]:
        """全テンプレートをリストアップする."""
        raise NotImplementedError()

    @abstractmethod
    def list_versions(self, name: str) -> list[str]:
        """指定された名前のテンプレートの全バージョンをリストアップする."""
        raise NotImplementedError()



"""各種定義クラス."""
from pydantic import BaseModel


class Property(BaseModel):
    """プロパティクラス."""

    name: str


class ObjectProperty(Property):
    """オブジェクトプロパティクラス."""

    properties: list[Property]


class ArrayProperty(Property):
    """配列プロパティクラス."""

    items: Property


class StringProperty(Property):
    """文字列プロパティクラス."""

    description: str


class Function(BaseModel):
    """関数クラス."""

    name: str
    description: str
    properties: list[Property]


class FunctionCallingConfig(BaseModel):
    """関数呼び出し設定クラス."""

    functions: list[Function]
    function_call: str | None = None


class Message(BaseModel):
    """メッセージクラス."""

    role: str
    content_type: str = 'text'
    content: str


class PromptTemplate(BaseModel):
    """プロンプトテンプレートクラス."""

    name: str
    description: str
    system: str | None = None
    messages: list[Message]
    parameters: dict[str, str]
    function_calling_config: FunctionCallingConfig | None = None
    version: str = '1.0'


class BuiltPrompt:
    """適用済みプロンプトクラス."""

    def __init__(self, system: str | None, messages: list[Message], function_calling_config: FunctionCallingConfig | None) -> None:
        """コンストラクタ."""
        self.system = system
        self.messages = messages
        self.function_calling_config = function_calling_config



"""AWS Bedrockモジュール."""
import boto3

from .adapter import LLMAdapter
from ..core.defines import BuiltPrompt
from ..core.defines import ObjectProperty
from ..core.defines import ArrayProperty
from ..core.defines import StringProperty

runtime = boto3.client('bedrock-runtime')


class BedrockClaude3Haiku(LLMAdapter):
    """AWS Bedrockとの通信を行うためのクラス."""

    def __init__(self) -> None:
        """コンストラクタ."""
        # FIXME その他パラメータをどのように設定するか？
        pass

    def _build_properties(self, properties) -> dict:
        """プロパティを構築する."""
        props = {}
        for prop in properties:
            if isinstance(prop, ObjectProperty):
                props[prop.name] = {
                    'type': 'object',
                    'properties': self._build_properties(prop.properties)
                }
            elif isinstance(prop, ArrayProperty):
                props[prop.name] = {
                    'type': 'array',
                    'items': self._build_properties([prop.items])
                }
            elif isinstance(prop, StringProperty):
                props[prop.name] = {
                    'type': 'string',
                    'description': prop.description
                }
        return props

    def _build_function(self, function) -> dict:
        """関数を構築する."""
        return {
            'toolSpec': {
                'name': function.name,
                'description': function.description,
                'inputSchema': {
                    'json': {
                        'type': 'object',
                        'properties': self._build_properties(function.properties)
                    }
                }
            }
        }

    def _build_tool_config(self, function_calling_config) -> dict:
        """ツール設定を構築する."""
        config = {
            'tools': [self._build_function(func) for func in function_calling_config.functions]
        }
        if function_calling_config.function_call:
            config['toolChoice'] = {
                'tool': {
                    'name': function_calling_config.function_call
                }
            }
        return config

    def call_llm(self, prompt: BuiltPrompt) -> tuple[str, int, int]:
        """Bedrockを呼び出す."""
        # FIXME 画像などのデータに対応する必要がある
        data = {
            'modelId': 'anthropic.claude-3-haiku-20240307-v1:0',
            'messages': [{
                'role': message.role,
                'content': [{
                    message.content_type: message.content
                }]
            } for message in prompt.messages],
        }
        if prompt.system:
            data['system'] = [{
                'text': prompt.system
            }]
        if prompt.function_calling_config:
            data['toolConfig'] = self._build_tool_config(prompt.function_calling_config)

        resp = runtime.converse(**data)
        if prompt.function_calling_config:
            output = resp['output']['message']['content'][0]['toolUse']['input']
        else:
            output = resp['output']['message']['content'][0]['text']

        input_tokens = resp['usage']['inputTokens']
        output_tokens = resp['usage']['outputTokens']
        return output, input_tokens, output_tokens
